# Generated by Django 3.2.18 on 2023-05-15 12:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0013_auto_20230317_1534'),
        ('tools', '0005_items_services_rights'),
    ]

    operations = [
        migrations.AddField(
            model_name='extract',
            name='health_facility',
            field=models.ForeignKey(db_column='HFID', on_delete=django.db.models.deletion.DO_NOTHING, to='location.healthfacility'),
        ),
        migrations.AddField(
            model_name='extract',
            name='location',
            field=models.ForeignKey(db_column='LocationId', on_delete=django.db.models.deletion.DO_NOTHING, to='location.location'),
        ),
    ]



import logging
import os
import tempfile
import xml.etree.ElementTree as ET

from core.models import Officer
from core.utils import filter_validity
from django.core.exceptions import PermissionDenied
from django.db.models.query_utils import Q
from django.http import HttpResponse
from django.http.response import FileResponse, JsonResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from django.views.decorators.http import require_http_methods
from import_export.resources import ModelResource
from tablib import Dataset

from location.models import HealthFacility, Location
from medical.models import Diagnosis, Item, Service
from location.apps import DEFAULT_CFG as LOCATION_DEFAULT_CFG
from rest_framework.decorators import api_view, permission_classes, renderer_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from . import serializers, services, utils
from .apps import ToolsConfig
from .resources import ItemResource, ServiceResource
from .services import return_upload_result_json

logger = logging.getLogger(__name__)


def checkUserWithRights(rights):
    class UserWithRights(IsAuthenticated):
        def has_permission(self, request, view):
            return super().has_permission(request, view) and request.user.has_perms(
                rights
            )

    return UserWithRights


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_locations_perms,
        )
    ]
)
@renderer_classes([serializers.LocationsXMLRenderer])
def download_locations(request):
    data = {"Regions": [], "Districts": [], "Municipalities": [], "Villages": []}

    for region in Location.objects.filter(~Q(code="FR"), type="R", *filter_validity()):
        data["Regions"].append(serializers.format_location(region))
        children = Location.objects.children(region.id).filter(validity_to=None)
        for child in children:
            if child.type == "D":
                data["Districts"].append(serializers.format_location(child))
            elif child.type == LOCATION_DEFAULT_CFG['location_types'][2]:
                data["Municipalities"].append(serializers.format_location(child))
            elif child.type == "V":
                data["Villages"].append(serializers.format_location(child))

    return Response(
        data,
        content_type="text/xml",
        headers={"Content-Disposition": "attachment; filename=locations.xml"},
    )


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_locations_perms,
        )
    ]
)
def upload_locations(request):
    serializer = serializers.UploadSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    dry_run = serializer.validated_data.get("dry_run")
    strategy = serializer.validated_data.get("strategy")

    try:
        logger.info(f"Uploading locations (dry_run={dry_run}, strategy={strategy})...")
        xml = utils.sanitize_xml(file)
        result = services.upload_locations(
            request.user, xml=xml, strategy=strategy, dry_run=dry_run
        )
        logger.info(f"Locations upload completed: {result}")
        return return_upload_result_json(xml_result=result)
    except ET.ParseError as exc:
        logger.error(exc)
        return Response(
            {
                "success": False,
                "error": "Malformed XML",
            }
        )


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_health_facilities_perms,
        )
    ]
)
@renderer_classes([serializers.HealthFacilitiesXMLRenderer])
def download_health_facilities(request):
    queryset = HealthFacility.objects.filter(*filter_validity())
    data = {
        "health_facility_details": [
            serializers.format_health_facility(hf) for hf in queryset
        ]
    }
    return Response(
        data=data,
        content_type="text/xml",
        headers={"Content-Disposition": "attachment; filename=health_facilities.xml"},
    )


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_health_facilities_perms,
        )
    ]
)
def upload_health_facilities(request):
    serializer = serializers.UploadSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    dry_run = serializer.validated_data.get("dry_run")
    strategy = serializer.validated_data.get("strategy")

    try:
        logger.info(
            f"Uploading health facilities (dry_run={dry_run}, strategy={strategy})..."
        )
        xml = utils.sanitize_xml(file)
        result = services.upload_health_facilities(
            request.user, xml=xml, strategy=strategy, dry_run=dry_run
        )
        logger.info(f"Health facilities upload completed: {result}")
        return return_upload_result_json(xml_result=result)
    except ET.ParseError as exc:
        logger.error(exc)
        return Response(
            {
                "success": False,
                "error": "Malformed XML",
            }
        )


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_diagnoses_perms,
        )
    ]
)
@renderer_classes([serializers.DiagnosesXMLRenderer])
def download_diagnoses(request):
    queryset = Diagnosis.objects.filter(*filter_validity())
    data = [serializers.format_diagnosis(hf) for hf in queryset]
    return Response(
        data=data,
        headers={"Content-Disposition": "attachment; filename=diagnoses.xml"},
    )


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_items_perms,
        )
    ]
)
@renderer_classes([serializers.ItemsXMLRenderer])
def download_items(request):
    """Downloads all medical.Item objects in an XML file.

    The list of exported fields can be modified in serializers.format_items().

    Calling this function is restricted to some users, see ToolsConfig.

    Parameters
    ----------
    request : rest_framework.request.Request
        The request that is asking to download items.

    Returns
    ----------
    rest_framework.response.Response
        The requested data in an XML file.

    """
    queryset = Item.objects.filter(*filter_validity())
    data = [serializers.format_items(item) for item in queryset]
    return Response(
        data=data,
        headers={"Content-Disposition": "attachment; filename=items.xml"},
    )


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_services_perms,
        )
    ]
)
@renderer_classes([serializers.ServicesXMLRenderer])
def download_services(request):
    queryset = Service.objects.filter(*filter_validity())
    data = [serializers.format_services(service) for service in queryset]
    return Response(
        data=data,
        headers={"Content-Disposition": "attachment; filename=services.xml"},
    )


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_diagnoses_perms,
        )
    ]
)
def upload_diagnoses(request):
    serializer = serializers.DeletableUploadSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    dry_run = serializer.validated_data.get("dry_run")
    strategy = serializer.validated_data.get("strategy")

    try:
        logger.info(f"Uploading diagnoses (dry_run={dry_run}, strategy={strategy})...")
        xml = utils.sanitize_xml(file)
        result = services.upload_diagnoses(
            request.user, xml=xml, strategy=strategy, dry_run=dry_run
        )
        logger.info(f"Diagnoses upload completed: {result}")
        return return_upload_result_json(xml_result=result)
    except ET.ParseError as exc:
        logger.error(exc)
        return Response(
            {
                "success": False,
                "error": "Malformed XML",
            }
        )


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_items_perms,
        )
    ]
)
def upload_items(request):
    """Uploads an XML file containing medical.Item entries.

    Calling this function is restricted to some users, see ToolsConfig.

    Parameters
    ----------
    request : rest_framework.request.Request
        The request containing all data.

    Returns
    ------
    JsonResponse
        Represents the number of entries received, with the number of created/updated/deleted
        medical.Item objects, as well as the list of errors that have occurred while each entry was processed.

    Raises
    ------
    ET.ParseError
        If the structure of the XML file is invalid.
    """
    serializer = serializers.DeletableUploadSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    dry_run = serializer.validated_data.get("dry_run")
    strategy = serializer.validated_data.get("strategy")

    try:
        logger.info("Uploading medical items (dry_run=%s, strategy=%s)...", dry_run, strategy)
        xml = utils.sanitize_xml(file)
        result = services.upload_items(
            request.user, xml=xml, strategy=strategy, dry_run=dry_run
        )
        logger.info("Medical items upload completed: %s", result)
        return return_upload_result_json(xml_result=result)
    except ET.ParseError as exc:
        logger.error(exc)
        return Response(
            {
                "success": False,
                "error": "Malformed XML",
            }
        )


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_services_perms,
        )
    ]
)
def upload_services(request):
    """Uploads an XML file containing medical.Service entries.

    Calling this function is restricted to some users, see ToolsConfig.

    Parameters
    ----------
    request : rest_framework.request.Request
        The request containing all data.

    Returns
    ------
    JsonResponse
        Represents the number of entries received, with the number of created/updated/deleted
        medical.Service objects, as well as the list of errors that have occurred while each entry was processed.

    Raises
    ------
    ET.ParseError
        If the structure of the XML file is invalid.
    """
    serializer = serializers.DeletableUploadSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    dry_run = serializer.validated_data.get("dry_run")
    strategy = serializer.validated_data.get("strategy")

    try:
        logger.info("Uploading medical services (dry_run=%s, strategy=%s)...", dry_run, strategy)
        xml = utils.sanitize_xml(file)
        result = services.upload_services(
            request.user, xml=xml, strategy=strategy, dry_run=dry_run
        )
        logger.info("Medical services upload completed: %s", result)
        return return_upload_result_json(xml_result=result)
    except ET.ParseError as exc:
        logger.error(exc)
        return Response(
            {
                "success": False,
                "error": "Malformed XML",
            }
        )


@api_view(["GET"])
def download_master_data(request):
    if not request.user.has_perms(ToolsConfig.extracts_master_data_perms):
        raise PermissionDenied(_("unauthorized"))

    export_file = services.create_master_data_export(request.user)

    response = FileResponse(
        open(export_file.name, "rb"),
        as_attachment=True,
        filename=os.path.basename(export_file.name),
        content_type="application/zip",
    )
    return response


@api_view(["GET"])
def download_phone_extract(request):
    if not request.user.has_perms(ToolsConfig.extracts_phone_extract_perms):
        raise PermissionDenied(_("unauthorized"))

    location_id = request.GET.get("location")
    with_insuree = request.GET.get("with_insuree", False)

    if not location_id:
        return Response(status=400, data={"error": "Location must be provided."})

    extract = services.create_phone_extract(request.user, location_id, with_insuree)

    return FileResponse(
        extract.stored_file,
        as_attachment=True,
        filename=os.path.basename(extract.stored_file.name),
    )


@api_view(["GET"])
def download_feedbacks(request):
    if not request.user.has_perms(ToolsConfig.extracts_officer_feedbacks_perms):
        raise PermissionDenied(_("unauthorized"))

    officer_id = request.GET.get("officer_id")
    officer = get_object_or_404(Officer, id=officer_id, *filter_validity())

    export_file = services.create_officer_feedbacks_export(request.user, officer)

    response = FileResponse(
        open(export_file.name, "rb"),
        as_attachment=True,
        filename=os.path.basename(export_file.name),
        content_type="application/zip",
    )

    return response

@api_view(["GET"])
def download_renewals(request):
    if not request.user.has_perms(ToolsConfig.extracts_officer_renewals_perms):
        raise PermissionDenied(_("unauthorized"))

    officer_id = request.GET.get("officer_id")
    officer = get_object_or_404(Officer, id=officer_id, *filter_validity())

    export_file = services.create_officer_renewals_export(request.user, officer)
    response = FileResponse(
        open(export_file.name, "rb"),
        as_attachment=True,
        filename=os.path.basename(export_file.name),
        content_type="application/zip",
    )

    return response


@require_http_methods(["POST"])
def upload_claims(request):
    if not request.user.has_perms(ToolsConfig.extracts_upload_claims_perms):
        raise PermissionDenied(_("unauthorized"))

    if not request.FILES:
        return JsonResponse({"error": "No file provided"}, status=400)

    errors = []
    for filename in request.FILES:
        try:
            logger.info(f"Processing claim in {filename}")
            xml = utils.sanitize_xml(request.FILES[filename])
            services.upload_claim(request.user, xml)
        except (utils.ParseError, services.InvalidXMLError) as exc:
            logger.exception(exc)
            errors.append(f"File '{filename}' is not a valid XML")
            continue
        except Exception as exc:
            logger.exception(exc)
            errors.append("An unknown error occurred.")
            continue

    return JsonResponse({"success": len(errors) == 0, "errors": errors})


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_locations_perms,
        )
    ]
)
# @require_http_methods(["POST"])
def upload_enrollments(request):
    if not request.user.has_perms(ToolsConfig.extracts_upload_claims_perms):  # TODO enrollment perms
        raise PermissionDenied(_("unauthorized"))

    return _process_upload(request, services.upload_enrollments)


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_locations_perms,
        )
    ]
)
# @require_http_methods(["POST"])
def upload_renewals(request):
    if not request.user.has_perms(ToolsConfig.extracts_upload_claims_perms):  # TODO enrollment perms
        raise PermissionDenied(_("unauthorized"))

    return _process_upload(request, services.upload_renewals)


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_locations_perms,
        )
    ]
)
# @require_http_methods(["POST"])
def upload_feedbacks(request):
    if not request.user.has_perms(ToolsConfig.extracts_upload_claims_perms):  # TODO enrollment perms
        raise PermissionDenied(_("unauthorized"))

    return _process_upload(request, services.upload_feedbacks)


def _process_upload(request, process_method):
    if not request.FILES:
        return JsonResponse({"error": "No file provided"}, status=400)

    errors = []
    for filename, uploaded_file in request.FILES.items():
        try:
            # The file is often in memory, the ZIP needs it as a real file path
            with tempfile.NamedTemporaryFile(mode='wb+', suffix=".zip") as tmp_file:
                logger.info(f"Processing renewals in {str(uploaded_file)}")
                for chunk in uploaded_file.chunks():
                    tmp_file.write(chunk)
                tmp_file.seek(0)
                process_method(tmp_file.name, request.user)
        except (utils.ParseError, services.InvalidXMLError) as exc:
            logger.exception(exc)
            errors.append(f"File '{filename}' is not a valid XML")
            continue
        except Exception as exc:
            logger.exception(exc)
            errors.append("An unknown error occurred.")
            return JsonResponse({"error": str(exc)}, status=500)

    return JsonResponse({"success": len(errors) == 0, "errors": errors})


# List of supported import/export formats so far
XLS = "xls"
XLSX = "xlsx"
CSV = "csv"
JSON = "json"
SUPPORTED_FORMATS = [XLS, XLSX, CSV, JSON]

# other types: https://stackoverflow.com/a/50860387
CONTENT_TYPES = {
    XLS: "application/vnd.ms-excel",
    CSV: "text/csv",
    JSON: "application/json",
    XLSX: "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
}


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_items_perms,
        )
    ]
)
def export_items(request):
    export_format = request.GET.get("file_format", "unknown")
    if export_format in SUPPORTED_FORMATS:
        return process_export_items(request.user.id_for_audit, export_format)
    else:
        return JsonResponse({"error": "Unknown export format."}, status=400)


def process_export_items(user_id, data_type):
    logger.info("User (audit id %s) requested export of medical items in %s", user_id, data_type)
    item_resource = ItemResource(user_id)
    query_set = Item.objects.filter(*filter_validity()).order_by("code")
    dataset = item_resource.export(query_set)
    datasets = {
        XLS: dataset.xls,
        CSV: dataset.csv,
        JSON: dataset.json,
        XLSX: dataset.xlsx,
    }
    response = HttpResponse(datasets[data_type], content_type=CONTENT_TYPES[data_type])
    response['Content-Disposition'] = f'attachment; filename="items.{data_type}"'
    return response


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_items_perms,
        )
    ]
)
def import_items(request):
    serializer = serializers.FileSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    user_id = request.user.id_for_audit
    logger.info("User (audit id %s) requested import of medical items", user_id)

    item_resource = ItemResource(user_id)
    dataset = Dataset()

    if CONTENT_TYPES[CSV] == file.content_type:
        dataset.load(file.read().decode(), format="csv")  # the CSV file must be read differently, otherwise it crashes
    else:
        dataset.load(file.read())

    return process_import_items_services(item_resource, dataset)


@api_view(["GET"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_services_perms,
        )
    ]
)
def export_services(request):
    export_format = request.GET.get("file_format", "unknown")
    if export_format in SUPPORTED_FORMATS:
        return process_export_services(request.user.id_for_audit, export_format)
    else:
        return JsonResponse({"error": "Unknown export format."}, status=400)


def process_export_services(user_id, data_type):
    logger.info("User (audit id %s) requested export of medical services in %s", user_id, data_type)
    service_resource = ServiceResource(user_id)
    query_set = Service.objects.filter(*filter_validity()).order_by("code")
    dataset = service_resource.export(query_set)
    datasets = {
        XLS: dataset.xls,
        CSV: dataset.csv,
        JSON: dataset.json,
        XLSX: dataset.xlsx,
    }
    response = HttpResponse(datasets[data_type], content_type=CONTENT_TYPES[data_type])
    response['Content-Disposition'] = f'attachment; filename="services.{data_type}"'
    return response


@api_view(["POST"])
@permission_classes(
    [
        checkUserWithRights(
            ToolsConfig.registers_services_perms,
        )
    ]
)
def import_services(request):
    serializer = serializers.FileSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    file = serializer.validated_data.get("file")
    user_id = request.user.id_for_audit
    logger.info("User (audit id %s) requested import of medical services", user_id)

    service_resource = ServiceResource(user_id)
    dataset = Dataset()

    if CONTENT_TYPES[CSV] == file.content_type:
        dataset.load(file.read().decode(), format="csv")  # the CSV file must be read differently, otherwise it crashes
    else:
        dataset.load(file.read())

    return process_import_items_services(service_resource, dataset)


def process_import_items_services(resource: ModelResource, dataset: Dataset):
    result = resource.import_data(dataset, dry_run=True)  # Test the data import

    success = True
    errors = []
    logger.info("Import results: total rows received=%s - detail=%s", result.total_rows, result.totals)

    if not result.has_errors() and not result.has_validation_errors():
        resource.import_data(dataset, dry_run=False)  # Actually import now
    else:
        success = False
        if result.has_validation_errors():
            for invalid_row in result.invalid_rows:
                logger.error("Invalid row n°%s - %s", invalid_row.number, invalid_row.error.message)
                errors.append(f"row ({invalid_row.number}) - {invalid_row.error.message}")
        if result.has_errors():
            for index, row_error in result.row_errors():
                logger.error("Error row n°%s", index)
                for error in row_error:
                    logger.error("Traceback: %s", error.traceback)
                    errors.append(f"row ({index}) - {error.error}")

    logger.info("End of import process")
    return return_upload_result_json(success=success, other_types_result=result, other_types_errors=errors)



# Generated by Django 3.2.16 on 2022-12-07 17:25
from core.utils import insert_role_right_for_system
from django.db import migrations


def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 131007)  # items
    insert_role_right_for_system(64, 131008)  # items
    insert_role_right_for_system(64, 131009)  # services
    insert_role_right_for_system(64, 131010)  # services


class Migration(migrations.Migration):

    dependencies = [
        ('tools', '0004_registers_right_for_scheme_admin'),
    ]

    operations = [
        migrations.RunPython(add_rights),
    ]



from core import PATIENT_CATEGORY_MASK_MALE, PATIENT_CATEGORY_MASK_FEMALE, PATIENT_CATEGORY_MASK_ADULT, \
    PATIENT_CATEGORY_MASK_MINOR, filter_validity
from import_export import resources, fields
from import_export.fields import Field
from import_export.widgets import IntegerWidget, BooleanWidget

from medical.models import Item, Service
from tools.services import validate_imported_item_row, validate_imported_service_row


def process_imported_patient_categories(row):
    # Transform the patient category data
    adult_cat = int(row["adult_cat"])
    minor_cat = int(row["minor_cat"])
    female_cat = int(row["female_cat"])
    male_cat = int(row["male_cat"])

    category = 0
    if male_cat:
        category = category | PATIENT_CATEGORY_MASK_MALE
    if female_cat:
        category = category | PATIENT_CATEGORY_MASK_FEMALE
    if adult_cat:
        category = category | PATIENT_CATEGORY_MASK_ADULT
    if minor_cat:
        category = category | PATIENT_CATEGORY_MASK_MINOR

    # Remove the now useless fields
    row.pop("adult_cat")
    row.pop("minor_cat")
    row.pop("female_cat")
    row.pop("male_cat")

    # Add the merged patient category value
    row["patient_category"] = category


class ItemServiceResource(resources.ModelResource):
    user_id = -1  # used for the audit_user_id field

    # These 4 fields are columns added for the export
    male_cat = Field(readonly=True)
    female_cat = Field(readonly=True)
    adult_cat = Field(readonly=True)
    minor_cat = Field(readonly=True)

    # This field is added to give users the possibility of deleting data with an upload
    delete = Field(widget=BooleanWidget())

    class Meta:
        # You can exclude some fields that will not be exported
        exclude = ('patient_category',)

        # You can specify which column should be used for data ID during import (by default = 'id')
        import_id_fields = ('code',)

    def __init__(self, given_id):
        super().__init__()
        self.user_id = given_id

    # The 4 dehydrate_xxx_cat methods generate the data that is going to be put in the xxx_cat columns
    def dehydrate_male_cat(self, item_service):
        result = item_service.patient_category & PATIENT_CATEGORY_MASK_MALE
        return result if not result else 1

    def dehydrate_female_cat(self, item_service):
        result = item_service.patient_category & PATIENT_CATEGORY_MASK_FEMALE
        return result if not result else 1

    def dehydrate_adult_cat(self, item_service):
        result = item_service.patient_category & PATIENT_CATEGORY_MASK_ADULT
        return result if not result else 1

    def dehydrate_minor_cat(self, item_service):
        result = item_service.patient_category & PATIENT_CATEGORY_MASK_MINOR
        return result if not result else 1

    # This method is called once before importing data
    # This is used to add the two mandatory fields that are required for creating a medical.Item
    # If self.fields do not have a fields.Field, with the column_name set up, then these columns are ignored during import
    def before_import(self, dataset, using_transactions, dry_run, **kwargs):
        if "patient_category" not in self.fields:
            self.fields["patient_category"] = fields.Field(attribute='patient_category', column_name="patient_category",
                                                           saves_null_values=False,
                                                           widget=IntegerWidget())
        if "audit_user_id" not in self.fields:
            self.fields["audit_user_id"] = fields.Field(attribute='audit_user_id', column_name="audit_user_id",
                                                        saves_null_values=False,
                                                        widget=IntegerWidget())

    # This method is called when the user flags a row to be deleted (the "delete" column value is '1')
    def for_delete(self, row, instance):
        if "delete" in row:
            return self.fields['delete'].clean(row)


# This class is responsible for customizing the import and export processes
class ItemResource(ItemServiceResource):

    class Meta:
        model = Item

        # These are the fields that are going to get exported/
        fields = ('code', 'name', 'type', 'package', 'price', 'quantity',
                  'care_type', 'frequency', 'patient_category')

        # You can customize the order for exports, but this order is also used during upload
        # (to know which fields will be there, instead of reading the headers)
        # export_order = ('code', 'name', 'type', 'package', 'price', 'quantity',
        #                 'care_type', 'frequency', 'male_cat', 'female_cat', 'adult_cat', 'minor_cat')

    # This method is called once for each row during import
    # This is where you can do some data validation/modification + add the missing data
    def before_import_row(self, row, **kwargs):
        row["type"] = row["type"].upper()
        row["care_type"] = row["care_type"].upper()
        validate_imported_item_row(row)
        process_imported_patient_categories(row)
        row["audit_user_id"] = self.user_id
        return row

    # This method is overridden in order to define which data is valid during import.
    def get_queryset(self):
        return Item.objects.filter(*filter_validity())


class ServiceResource(ItemServiceResource):

    class Meta:
        model = Service

        # These are the fields that are going to get exported/
        fields = ('code', 'name', 'type', 'level', 'price', 'category',
                  'care_type', 'frequency', 'patient_category')

    # This method is called once for each row during import
    # This is where you can do some data validation/modification + add the missing data
    def before_import_row(self, row, **kwargs):
        row["type"] = row["type"].upper()
        row["care_type"] = row["care_type"].upper()
        row["level"] = row["level"].upper()
        if row["category"] is not None and len(row["category"]) > 0:  # optional field
            row["category"] = row["category"].upper()
        validate_imported_service_row(row)
        process_imported_patient_categories(row)
        row["audit_user_id"] = self.user_id
        return row

    # This method is overridden in order to define which data is valid during import.
    def get_queryset(self):
        return Service.objects.filter(*filter_validity())



from unittest.mock import patch
from tempfile import TemporaryFile

from core import filter_validity
from core.test_helpers import create_test_interactive_user
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase

from medical.models import Item
from medical.test_helpers import create_test_item
from tools.services import parse_xml_items, UploadResult, upload_items, parse_optional_item_fields
from tools.constants import STRATEGY_INSERT, STRATEGY_UPDATE, STRATEGY_INSERT_UPDATE, STRATEGY_INSERT_UPDATE_DELETE
from xml.etree import ElementTree


class UploadItemsParseXMLItemsTestCase(TestCase):

    def test_parse_xml_item_fields_no_code(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemName>Item code missing</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>42.00</ItemPrice>
                    <ItemCareType>B</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemQuantity/>
                    <ItemPackage/>
                    <ItemFrequency/>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Item is missing one of", errors[0])

    def test_parse_xml_item_fields_no_name(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode>NO_NAME</ItemCode>
                    <ItemType>M</ItemType>
                    <ItemPrice>42.42</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Nice package</ItemPackage>
                    <ItemQuantity/>
                    <ItemFrequency>88</ItemFrequency>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Item is missing one of", errors[0])

    def test_parse_xml_item_fields_no_item_type(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode>NO_TYPE</ItemCode>
                    <ItemName>Item type missing</ItemName>
                    <ItemPrice>42.42</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Item is missing one of", errors[0])

    def test_parse_xml_item_fields_no_price(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode>NO_PRICE</ItemCode>
                    <ItemName>Item price missing</ItemName>
                    <ItemType>D</ItemType>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Something</ItemPackage>
                    <ItemQuantity>8.99</ItemQuantity>
                    <ItemFrequency/>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Item is missing one of", errors[0])

    def test_parse_xml_item_fields_no_care_type(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode>NO_CARE_TYPE</ItemCode>
                    <ItemName>Item care_type missing</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>1000 TABLETS</ItemPackage>
                    <ItemQuantity>42.44</ItemQuantity>
                    <ItemFrequency>2</ItemFrequency>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Item is missing one of", errors[0])

    def test_parse_xml_item_fields_various_price_errors(self):
        code_1 = "ERROR_PRICE_1"
        code_2 = "ERROR_PRICE_2"
        code_3 = "ERROR_PRICE_3"
        code_ok = "OK"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>{code_1}</ItemCode>
                    <ItemName>Error in price - wrong decimal separator</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>123,4</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemQuantity/>
                    <ItemPackage/>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_2}</ItemCode>
                    <ItemName>Error in price - currency symbol at the end</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>123.4€</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemQuantity/>
                    <ItemPackage/>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_3}</ItemCode>
                    <ItemName>Error in price - currency symbol as decimal separator</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>123€4</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemQuantity/>
                    <ItemPackage/>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_ok}</ItemCode>
                    <ItemName>Proper price, with a '.' for decimal separator and without currency symbol</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>123.45</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemQuantity/>
                    <ItemPackage/>
                    <ItemFrequency/>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertTrue(raw_items)
            self.assertEqual(len(raw_items), 1)
            self.assertEqual(raw_items[0]["code"], code_ok)
            self.assertEqual(raw_items[0]["price"], 123.45)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 3)

            self.assertIn("price is invalid", errors[0])
            self.assertIn(code_1, errors[0])

            self.assertIn("price is invalid", errors[1])
            self.assertIn(code_2, errors[1])

            self.assertIn("price is invalid", errors[2])
            self.assertIn(code_3, errors[2])

    def test_parse_xml_item_fields_error_repeated_code(self):
        code_1 = "CODE_1"
        code_2 = "CODE_2"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>{code_1}</ItemCode>
                    <ItemName>Item 1</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>0.45</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity/>
                    <ItemFrequency>5</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>{code_1}</ItemCode>
                    <ItemName>Item 1 w/ different values and same code</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>2.45</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>0</ItemFemaleCategory>
                    <ItemAdultCategory>0</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity/>
                    <ItemFrequency>5</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>{code_2}</ItemCode>
                    <ItemName>Item 2</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1000</ItemPrice>
                    <ItemCareType>B</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity/>
                    <ItemFrequency>5</ItemFrequency>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertTrue(raw_items)
            self.assertEqual(len(raw_items), 2)
            self.assertEqual(raw_items[0]["code"], code_1)
            self.assertEqual(raw_items[1]["code"], code_2)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("exists multiple times", errors[0])
            self.assertIn(code_1, errors[0])

    def test_parse_xml_item_fields_error_code_too_small(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode> </ItemCode>
                    <ItemName>Item code is a single space that will be trimmed</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity/>
                    <ItemFrequency>58</ItemFrequency>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("code is invalid", errors[0])
            self.assertIn("''", errors[0])

    def test_parse_xml_item_fields_error_code_too_long(self):
        long_boi = "THIS_CODE_IS_REALLY_LONG_WHY?"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>{long_boi}</ItemCode>
                    <ItemName>Item code is a single space that will be trimmed</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("code is invalid", errors[0])
            self.assertIn(long_boi, errors[0])

    def test_parse_xml_item_fields_error_name_too_small(self):
        xml = b"""
            <Items>
                <Item>
                    <ItemCode>CODE</ItemCode>
                    <ItemName> </ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>882.5</ItemQuantity>
                    <ItemFrequency/>
                </Item>
            </Items>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("name is invalid", errors[0])
            self.assertIn("''", errors[0])

    def test_parse_xml_item_fields_error_name_too_long(self):
        long_boi = "this name is really long, why? Why would anyone prepare a medical " \
                   "item with a name so long, it doesn't make any sense..."
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>CODE</ItemCode>
                    <ItemName>{long_boi}</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Pikachu</ItemPackage>
                    <ItemQuantity/>
                    <ItemFrequency/>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertFalse(raw_items)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("name is invalid", errors[0])
            self.assertIn(long_boi, errors[0])

    def test_parse_xml_item_fields_error_unknown_type(self):
        unknown_type = "O"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>M_MAJ</ItemCode>
                    <ItemName>Item of type M in capital letter</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>9.87</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>M_MIN</ItemCode>
                    <ItemName>Item of type M in small letter</ItemName>
                    <ItemType>m</ItemType>
                    <ItemPrice>8.76</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>D_MAJ</ItemCode>
                    <ItemName>Item of type D in capital letter</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>7.65</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>D_MIN</ItemCode>
                    <ItemName>Item of type D in small letter</ItemName>
                    <ItemType>d</ItemType>
                    <ItemPrice>6.54</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>O_MIN</ItemCode>
                    <ItemName>Item of unknown type</ItemName>
                    <ItemType>{unknown_type}</ItemType>
                    <ItemPrice>5.43</ItemPrice>
                    <ItemCareType>B</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertTrue(raw_items)
            self.assertEqual(len(raw_items), 4)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("type is invalid", errors[0])
            self.assertIn(unknown_type, errors[0])

    def test_parse_xml_item_fields_error_unknown_care_type(self):
        unknown_care_type = "K"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>I_MAJ</ItemCode>
                    <ItemName>Item of care_type I in capital letter</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>9.87</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Mew</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>I_MIN</ItemCode>
                    <ItemName>Item of care_type I in small letter</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>8.76</ItemPrice>
                    <ItemCareType>i</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Mewtwo</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>O_MAJ</ItemCode>
                    <ItemName>Item of care_type O in capital letter</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>7.65</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Snorlax</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>O_MIN</ItemCode>
                    <ItemName>Item of care_type O in small letter</ItemName>
                    <ItemType>m</ItemType>
                    <ItemPrice>6.54</ItemPrice>
                    <ItemCareType>o</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Chansey</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>B_MAJ</ItemCode>
                    <ItemName>Item of care_type B in capital letter</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>5.43</ItemPrice>
                    <ItemCareType>B</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Charizard</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>B_MIN</ItemCode>
                    <ItemName>Item of care_type B in small letter</ItemName>
                    <ItemType>m</ItemType>
                    <ItemPrice>4.32</ItemPrice>
                    <ItemCareType>b</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Machop</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
                <Item>
                    <ItemCode>K_MAJ</ItemCode>
                    <ItemName>Item of unknown care_type</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>3.21</ItemPrice>
                    <ItemCareType>{unknown_care_type}</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage>Tortank</ItemPackage>
                    <ItemQuantity>1.25</ItemQuantity>
                    <ItemFrequency>7</ItemFrequency>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertTrue(raw_items)
            self.assertEqual(len(raw_items), 6)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("care type is invalid", errors[0])
            self.assertIn(unknown_care_type, errors[0])

    def test_parse_xml_item_fields_mixed_errors_and_success(self):
        code_ok_1 = "CODE_1"
        code_ok_2 = "CODE_2"
        code_ok_3 = "CODE_3"

        long_name = "this name is really long, why? Why would anyone prepare a medical item " \
                    "with a name so long, it doesn't make any sense..."
        unknown_care_type = "K"
        xml = f"""
            <Items>
                <Item>
                    <ItemCode>CODE</ItemCode>
                    <ItemName>{long_name}</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>1234</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_ok_1}</ItemCode>
                    <ItemName>Valid item 1 - no error</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>9.87</ItemPrice>
                    <ItemCareType>I</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>K_MAJ</ItemCode>
                    <ItemName>Item of unknown care_type</ItemName>
                    <ItemType>M</ItemType>
                    <ItemPrice>3.21</ItemPrice>
                    <ItemCareType>{unknown_care_type}</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_ok_2}</ItemCode>
                    <ItemName>Valid item 2 - no error</ItemName>
                    <ItemType>d</ItemType>
                    <ItemPrice>555.55</ItemPrice>
                    <ItemCareType>b</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemCode>{code_ok_3}</ItemCode>
                    <ItemName>Valid item 3 - no error</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>489.54</ItemPrice>
                    <ItemCareType>O</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
                <Item>
                    <ItemName>Item without code</ItemName>
                    <ItemType>D</ItemType>
                    <ItemPrice>3700.88</ItemPrice>
                    <ItemCareType>o</ItemCareType>
                    <ItemMaleCategory>1</ItemMaleCategory>
                    <ItemFemaleCategory>1</ItemFemaleCategory>
                    <ItemAdultCategory>1</ItemAdultCategory>
                    <ItemMinorCategory>1</ItemMinorCategory>
                    <ItemPackage/>
                    <ItemQuantity>88.25</ItemQuantity>
                    <ItemFrequency/>
                </Item>
            </Items>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_items, errors = parse_xml_items(et)

            self.assertTrue(raw_items)
            self.assertEqual(len(raw_items), 3)
            self.assertIn(code_ok_1, raw_items[0]["code"])
            self.assertIn(code_ok_2, raw_items[1]["code"])
            self.assertIn(code_ok_3, raw_items[2]["code"])

            self.assertTrue(errors)
            self.assertEqual(len(errors), 3)
            self.assertIn("name is invalid", errors[0])
            self.assertIn(long_name, errors[0])
            self.assertIn("care type is invalid", errors[1])
            self.assertIn(unknown_care_type, errors[1])
            self.assertIn("Item is missing one of", errors[2])


class UploadItemsParseOptionalFieldsTestCase(TestCase):

    def test_parse_optional_item_fields_all_empty(self):
        xml = f"""
            <Item>
                <ItemQuantity/>
                <ItemPackage/>
                <ItemFrequency/>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "EMPTY")

            self.assertFalse(optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_all_filled(self):
        xml = f"""
            <Item>
                <ItemPackage>1000 TABLETS</ItemPackage>
                <ItemQuantity>42.00</ItemQuantity>
                <ItemFrequency>5</ItemFrequency>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "FILLED")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 3)
            self.assertIn("frequency", optional_fields)
            self.assertIn("quantity", optional_fields)
            self.assertIn("package", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_frq(self):
        xml = f"""
            <Item>
                <ItemPackage/>
                <ItemQuantity/>
                <ItemFrequency>5</ItemFrequency>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "1/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 1)
            self.assertIn("frequency", optional_fields)
            self.assertNotIn("quantity", optional_fields)
            self.assertNotIn("package", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_qty(self):
        xml = f"""
            <Item>
                <ItemPackage/>
                <ItemQuantity>88.25</ItemQuantity>
                <ItemFrequency/>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "1/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 1)
            self.assertIn("quantity", optional_fields)
            self.assertNotIn("frequency", optional_fields)
            self.assertNotIn("package", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_pkg(self):
        xml = f"""
            <Item>
                <ItemPackage>Great package</ItemPackage>
                <ItemQuantity/>
                <ItemFrequency/>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "1/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 1)
            self.assertIn("package", optional_fields)
            self.assertNotIn("quantity", optional_fields)
            self.assertNotIn("frequency", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_frq_qty(self):
        xml = f"""
            <Item>
                <ItemPackage/>
                <ItemQuantity>1.25</ItemQuantity>
                <ItemFrequency>7</ItemFrequency>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "2/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 2)
            self.assertIn("frequency", optional_fields)
            self.assertIn("quantity", optional_fields)
            self.assertNotIn("package", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_frq_pkg(self):
        xml = f"""
            <Item>
                <ItemPackage>Nice package</ItemPackage>
                <ItemQuantity/>
                <ItemFrequency>88</ItemFrequency>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "2/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 2)
            self.assertIn("frequency", optional_fields)
            self.assertIn("package", optional_fields)
            self.assertNotIn("quantity", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_filled_and_empty_qty_pkg(self):
        xml = f"""
            <Item>
                <ItemPackage>Nice package</ItemPackage>
                <ItemQuantity>8.99</ItemQuantity>
                <ItemFrequency/>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_item_fields(root, "2/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 2)
            self.assertIn("package", optional_fields)
            self.assertIn("quantity", optional_fields)
            self.assertNotIn("frequency", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_item_fields_error_quantity(self):
        xml = f"""
            <Item>
                <ItemPackage>1000 TABLETS</ItemPackage>
                <ItemQuantity>42ml</ItemQuantity>
                <ItemFrequency>2</ItemFrequency>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            code = "ERROR_QTY"
            optional_fields, error = parse_optional_item_fields(root, code)

            self.assertFalse(optional_fields)
            self.assertTrue(error)
            self.assertIn("quantity is invalid", error)
            self.assertIn(code, error)

    def test_parse_optional_item_fields_error_frequency(self):
        xml = f"""
            <Item>
                <ItemPackage>1000 TABLETS</ItemPackage>
                <ItemFrequency>5.55</ItemFrequency>
                <ItemQuantity>42.00</ItemQuantity>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            code = "ERROR_FRQ"
            optional_fields, error = parse_optional_item_fields(root, code)

            self.assertNotIn("frequency", optional_fields)
            self.assertTrue(error)
            self.assertIn("frequency is invalid", error)
            self.assertIn(code, error)

    def test_parse_optional_item_fields_error_package_too_small(self):
        xml = f"""
            <Item>
                <ItemPackage> </ItemPackage>
                <ItemFrequency>1</ItemFrequency>
                <ItemQuantity>42.00</ItemQuantity>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            code = "ERROR_PKG"
            optional_fields, error = parse_optional_item_fields(root, code)

            self.assertNotIn("package", optional_fields)
            self.assertTrue(error)
            self.assertIn("package is invalid", error)
            self.assertIn(code, error)

    def test_parse_optional_item_fields_error_package_too_long(self):
        long_boi = "this package is really long, why? Why would anyone prepare a medical item " \
                   "with a packaging so long that it doesn't fit in 255 characters, it doesn't " \
                   "make any sense... and still, i haven't currently reached 255 characters yet, so " \
                   "I am trying to get there and writing anything... please enjoy if you read these tests later :)"
        xml = f"""
            <Item>
                <ItemPackage>{long_boi}</ItemPackage>
                <ItemFrequency>12</ItemFrequency>
                <ItemQuantity>2.00</ItemQuantity>
            </Item>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            code = "ERROR_PKG"
            optional_fields, error = parse_optional_item_fields(root, code)

            self.assertNotIn("package", optional_fields)
            self.assertTrue(error)
            self.assertIn("package is invalid", error)
            self.assertIn(code, error)


class UploadItemsTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username="testItemsAdmin")

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT
        existing_code = "0001"
        raw_items = [
            {
                "code": "CODE_1",
                "name": "Valid item 1 - no error",
            },
            {
                "code": "CODE_2",
                "name": "Valid item 2 - no error",
            },
            {
                "code": existing_code,
                "name": "Invalid item - code already exists",
            }
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors

        expected_errors = [
            f"Item '{existing_code}' already exists"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=2,
            updated=0,
            deleted=0,
        )
        result = upload_items(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_update_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_UPDATE
        existing_code_1 = "0001"
        existing_code_2 = "0127"
        error_code = "ERROR"
        raw_items = [
            {
                "code": existing_code_1,
                "name": "Valid item 1 - no error",
            },
            {
                "code": error_code,
                "name": "Invalid item - code doesn't exists",
            },
            {
                "code": existing_code_2,
                "name": "Valid item 2 - no error",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors

        expected_errors = [
            f"Item '{error_code}' does not exist"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=0,
            updated=2,
            deleted=0,
        )
        result = upload_items(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert_update_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT_UPDATE
        raw_items = [
            {
                "code": "CODE_1",
                "name": "New item 1",
            },
            {
                "code": "CODE_2",
                "name": "New item 2",
            },
            {
                "code": "0001",
                "name": "Existing item 1",
            },
            {
                "code": "CODE_3",
                "name": "New item 3",
            },
            {
                "code": "0042",
                "name": "Existing item 2",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors

        expected = UploadResult(
            errors=errors,
            sent=5,
            created=3,
            updated=2,
            deleted=0,
        )
        result = upload_items(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert_update_delete_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT_UPDATE_DELETE
        raw_items = [
            {
                "code": "CODE_1",
                "name": "New item 1",
            },
            {
                "code": "CODE_2",
                "name": "New item 2",
            },
            {
                "code": "0001",
                "name": "Existing item 1",
            },
            {
                "code": "CODE_3",
                "name": "New item 3",
            },
            {
                "code": "0042",
                "name": "Existing item 2",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors

        before_total_items = Item.objects.filter(*filter_validity()).count()
        expected_deleted = before_total_items - 2

        expected = UploadResult(
            errors=errors,
            sent=5,
            created=3,
            updated=2,
            deleted=expected_deleted,
        )
        result = upload_items(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert(self, mock_parsing):
        # setup - preparing data that will be inserted
        dry_run = False
        strategy = STRATEGY_INSERT
        existing_code = "0001"
        new_code_1 = "CODE_1"
        new_code_2 = "CODE_2"
        raw_items = [
            {
                "code": new_code_1,
                "name": "Valid item 1 - no error",
                "type": "D",
                "price": 489.54,
                "care_type": "O",
                "patient_category": 15,
                "package": "package",
                "frequency": 5,
                "quantity": 1.2,
            },
            {
                "code": new_code_2,
                "name": "Valid item 2 - no error",
                "type": "D",
                "price": 499.54,
                "care_type": "O",
                "patient_category": 5,
                "package": "package",
                "frequency": 57,
                "quantity": 1.27,
            },
            {
                "code": existing_code,
                "name": "Invalid item - code already exists",
                "type": "D",
                "price": 599.54,
                "care_type": "B",
                "patient_category": 7,
                "package": "package",
                "frequency": 7,
                "quantity": 7,
            }
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors

        expected_errors = [
            f"Item '{existing_code}' already exists"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=2,
            updated=0,
            deleted=0,
        )
        total_items_before = Item.objects.filter(*filter_validity()).count()

        # Inserting
        result = upload_items(self.admin_user, "xml", strategy, dry_run)
        total_items_after = Item.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_items_before + 2, total_items_after)

        # Making sure the new items don't stay in the DB
        new_item_1 = Item.objects.get(code=new_code_1)
        new_item_1.delete()
        new_item_2 = Item.objects.get(code=new_code_2)
        new_item_2.delete()

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_update(self, mock_parsing):
        # setup - creating items that will be updated
        new_code_1 = "CODE_1"
        old_name_1 = "new item old name 0001"
        new_item_1_props = {
            "code": new_code_1,
            "name": old_name_1,
        }
        create_test_item(item_type="D", custom_props=new_item_1_props)

        new_code_2 = "CODE_2"
        old_name_2 = "new item old name 0002"
        new_item_2_props = {
            "code": new_code_2,
            "name": old_name_2,
        }
        create_test_item(item_type="D", custom_props=new_item_2_props)

        # setup - preparing values used for the update
        dry_run = False
        strategy = STRATEGY_UPDATE
        new_name_1 = "new item new name 0001"
        new_name_2 = "new item new name 0002"
        non_existing_code = "ERROR"
        raw_items = [
            {
                "code": new_code_1,
                "name": new_name_1,
                "type": "m",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "package": "package",
                "quantity": 7,
            },
            {
                "code": new_code_2,
                "name": new_name_2,
                "type": "m",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
                "package": "package",
                "quantity": 7,
            },
            {
                "code": non_existing_code,
                "name": "error - this can't be updated",
                "type": "m",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
            }
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors
        expected_errors = [
            f"Item '{non_existing_code}' does not exist"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=0,
            updated=2,
            deleted=0,
        )
        total_items_before = Item.objects.filter(*filter_validity()).count()

        # update
        result = upload_items(self.admin_user, "xml", strategy, dry_run)
        total_items_after = Item.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_items_before, total_items_after)

        # Making sure the names have been updated + deleting the new items to make sure they don't stay in the DB
        db_new_item_1 = Item.objects.get(code=new_code_1, validity_to=None)
        self.assertEqual(db_new_item_1.name, new_name_1)
        db_new_item_1.delete()

        db_new_item_2 = Item.objects.get(code=new_code_2, validity_to=None)
        self.assertEqual(db_new_item_2.name, new_name_2)
        db_new_item_2.delete()

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert_update(self, mock_parsing):
        # setup - creating items that will be updated
        update_code_1 = "U_1"
        old_name_1 = "update item old name 0001"
        update_item_1_props = {
            "code": update_code_1,
            "name": old_name_1,
        }
        create_test_item(item_type="D", custom_props=update_item_1_props)

        update_code_2 = "U_2"
        old_name_2 = "update item old name 0002"
        update_item_2_props = {
            "code": update_code_2,
            "name": old_name_2,
        }
        create_test_item(item_type="D", custom_props=update_item_2_props)

        # setup - preparing values used for the update
        dry_run = False
        strategy = STRATEGY_INSERT_UPDATE
        new_name_1 = "update item new name 0001"
        new_name_2 = "update item new name 0002"
        insert_code_1 = "I_1"
        insert_code_2 = "I_2"
        raw_items = [
            {
                "code": insert_code_1,
                "name": "insert item 1",
                "type": "m",
                "price": 65.54,
                "care_type": "b",
                "patient_category": 15,
                "package": "yes",
                "quantity": 7,
            },
            {
                "code": update_code_1,
                "name": new_name_1,
                "type": "m",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "package": "package",
                "quantity": 7,
                "frequency": 6,
            },
            {
                "code": update_code_2,
                "name": new_name_2,
                "type": "m",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
            },
            {
                "code": insert_code_2,
                "name": "insert item 2",
                "type": "m",
                "price": 60.54,
                "care_type": "b",
                "patient_category": 15,
                "quantity": 7,
            },
        ]
        errors = []
        mock_parsing.return_value = raw_items, errors
        expected = UploadResult(
            errors=errors,
            sent=4,
            created=2,
            updated=2,
            deleted=0,
        )
        total_items_before = Item.objects.filter(*filter_validity()).count()

        # insert-update
        result = upload_items(self.admin_user, "xml", strategy, dry_run)
        total_items_after = Item.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_items_before + 2, total_items_after)

        # Making sure the names have been updated + deleting the update items to make sure they don't stay in the DB
        db_update_item_1 = Item.objects.get(code=update_code_1, validity_to=None)
        self.assertEqual(db_update_item_1.name, new_name_1)
        db_update_item_1.delete()

        db_update_item_2 = Item.objects.get(code=update_code_2, validity_to=None)
        self.assertEqual(db_update_item_2.name, new_name_2)
        db_update_item_2.delete()

        # Also deleting the insert items to make sure they don't stay in the DB
        db_insert_item_1 = Item.objects.get(code=insert_code_1, validity_to=None)
        db_insert_item_1.delete()
        db_insert_item_2 = Item.objects.get(code=insert_code_2, validity_to=None)
        db_insert_item_2.delete()

    @patch("tools.services.parse_xml_items")
    def test_upload_items_multiple_insert_update_delete(self, mock_parsing):
        # setup - fetching initial DB items in order not to delete them
        all_items = Item.objects.filter(*filter_validity()).all()
        items_to_not_delete = []
        for item in all_items:
            item_dict = item.__dict__
            item_dict.pop("id")
            item_dict.pop("uuid")
            item_dict.pop("_state")
            item_dict.pop("validity_from")
            item_dict.pop("validity_to")
            item_dict.pop("legacy_id")
            items_to_not_delete.append(item_dict)

        # setup - creating items that will be updated & deleted
        update_code = "U_1"
        old_name = "update item old name 0001"
        update_item_props = {
            "code": update_code,
            "name": old_name,
        }
        create_test_item(item_type="D", custom_props=update_item_props)

        delete_code = "D_1"
        delete_item_props = {
            "code": delete_code,
            "name": "item do be deleted",
        }
        create_test_item(item_type="D", custom_props=delete_item_props)

        # setup - preparing values used for the tests
        dry_run = False
        strategy = STRATEGY_INSERT_UPDATE_DELETE
        update_new_name = "update item new name 0001"
        insert_code_1 = "I_1"
        insert_code_2 = "I_2"
        new_items = [
            {
                "code": insert_code_1,
                "name": "insert item 1",
                "type": "m",
                "price": 65.54,
                "care_type": "b",
                "patient_category": 15,
                "package": "yes",
                "quantity": 7,
            },
            {
                "code": update_code,
                "name": update_new_name,
                "type": "m",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "package": "package",
                "quantity": 7,
                "frequency": 6,
            },
            {
                "code": insert_code_2,
                "name": "insert item 2",
                "type": "m",
                "price": 60.54,
                "care_type": "b",
                "patient_category": 15,
                "quantity": 7,
            },
        ]
        raw_items = new_items + items_to_not_delete
        errors = []
        mock_parsing.return_value = raw_items, errors

        total_updated = len(items_to_not_delete) + 1
        total_sent = total_updated + 2
        expected = UploadResult(
            errors=errors,
            sent=total_sent,
            created=2,
            updated=total_updated,
            deleted=1,
        )
        total_items_before = Item.objects.filter(*filter_validity()).count()

        # insert update delete
        result = upload_items(self.admin_user, "xml", strategy, dry_run)
        total_items_after = Item.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        # 2 inserts and 1 deletion
        self.assertEqual(total_items_before + 1, total_items_after)

        with self.assertRaises(ObjectDoesNotExist):
            Item.objects.get(code=delete_code, validity_to=None)

        # Making sure the name has been updated + deleting the items to make sure they don't stay in the DB
        db_update_item = Item.objects.get(code=update_code, validity_to=None)
        self.assertEqual(db_update_item.name, update_new_name)
        db_update_item.delete()

        db_insert_item_1 = Item.objects.get(code=insert_code_1, validity_to=None)
        db_insert_item_1.delete()
        db_insert_item_2 = Item.objects.get(code=insert_code_2, validity_to=None)
        db_insert_item_2.delete()



from unittest.mock import patch
from tempfile import TemporaryFile

from core import filter_validity
from core.test_helpers import create_test_interactive_user
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase

from medical.models import Service
from medical.test_helpers import create_test_service
from tools.services import parse_xml_services, UploadResult, upload_services, parse_optional_service_fields
from tools.constants import STRATEGY_INSERT, STRATEGY_UPDATE, STRATEGY_INSERT_UPDATE, STRATEGY_INSERT_UPDATE_DELETE
from xml.etree import ElementTree


class UploadServicesParseXMLServicesTestCase(TestCase):

    def test_parse_xml_service_fields_no_code(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceName>Service code missing</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>S</ServiceLevel>
                    <ServicePrice>42.00</ServicePrice>
                    <ServiceCareType>B</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_no_name(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>NO_NAME</ServiceCode>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>v</ServiceLevel>
                    <ServicePrice>42.42</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>a</ServiceCategory>
                    <ServiceFrequency>88</ServiceFrequency>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_no_service_type(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>NO_TYPE</ServiceCode>
                    <ServiceName>Service type missing</ServiceName>
                    <ServiceLevel>C</ServiceLevel>
                    <ServicePrice>42.42</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_no_level(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>NO_LEVEL</ServiceCode>
                    <ServiceName>Service level missing</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServicePrice>42.42</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>V</ServiceCategory>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_no_price(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>NO_PRICE</ServiceCode>
                    <ServiceName>Service price missing</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>h</ServiceLevel>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>V</ServiceCategory>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_no_care_type(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>NO_CARE_TYPE</ServiceCode>
                    <ServiceName>Service care_type missing</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>V</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>c</ServiceCategory>
                    <ServiceFrequency>2</ServiceFrequency>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("Service is missing one of", errors[0])

    def test_parse_xml_service_fields_various_price_errors(self):
        code_1 = "ERROR_PRICE_1"
        code_2 = "ERROR_PRICE_2"
        code_3 = "ERROR_PRICE_3"
        code_ok = "OK"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>{code_1}</ServiceCode>
                    <ServiceName>Error in price - wrong decimal separator</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>s</ServiceLevel>
                    <ServicePrice>123,4</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_2}</ServiceCode>
                    <ServiceName>Error in price - currency symbol at the end</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>D</ServiceLevel>
                    <ServicePrice>123.4€</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_3}</ServiceCode>
                    <ServiceName>Error in price - currency symbol as decimal separator</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>H</ServiceLevel>
                    <ServicePrice>123€4</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_ok}</ServiceCode>
                    <ServiceName>Proper price, with a '.' for decimal separator and without currency symbol</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>h</ServiceLevel>
                    <ServicePrice>123.45</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertTrue(raw_services)
            self.assertEqual(len(raw_services), 1)
            self.assertEqual(raw_services[0]["code"], code_ok)
            self.assertEqual(raw_services[0]["price"], 123.45)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 3)

            self.assertIn("price is invalid", errors[0])
            self.assertIn(code_1, errors[0])

            self.assertIn("price is invalid", errors[1])
            self.assertIn(code_2, errors[1])

            self.assertIn("price is invalid", errors[2])
            self.assertIn(code_3, errors[2])

    def test_parse_xml_service_fields_error_repeated_code(self):
        code_1 = "CODE_1"
        code_2 = "CODE_2"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>{code_1}</ServiceCode>
                    <ServiceName>Service 1</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>H</ServiceLevel>
                    <ServicePrice>0.45</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>5</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>{code_1}</ServiceCode>
                    <ServiceName>Service 1 w/ different values and same code</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>h</ServiceLevel>
                    <ServicePrice>2.45</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>0</ServiceFemaleCategory>
                    <ServiceAdultCategory>0</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>5</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>{code_2}</ServiceCode>
                    <ServiceName>Service 2</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>V</ServiceLevel>
                    <ServicePrice>1000</ServicePrice>
                    <ServiceCareType>B</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>5</ServiceFrequency>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertTrue(raw_services)
            self.assertEqual(len(raw_services), 2)
            self.assertEqual(raw_services[0]["code"], code_1)
            self.assertEqual(raw_services[1]["code"], code_2)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("exists multiple times", errors[0])
            self.assertIn(code_1, errors[0])

    def test_parse_xml_service_fields_error_code_too_small(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode> </ServiceCode>
                    <ServiceName>Service code is a single space that will be trimmed</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>s</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>58</ServiceFrequency>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("code is invalid", errors[0])
            self.assertIn("''", errors[0])

    def test_parse_xml_service_fields_error_code_too_long(self):
        long_boi = "THIS_CODE_IS_REALLY_LONG_WHY?"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>{long_boi}</ServiceCode>
                    <ServiceName>Service code is a single space that will be trimmed</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>V</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("code is invalid", errors[0])
            self.assertIn(long_boi, errors[0])

    def test_parse_xml_service_fields_error_name_too_small(self):
        xml = b"""
            <Services>
                <Service>
                    <ServiceCode>CODE</ServiceCode>
                    <ServiceName> </ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>v</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("name is invalid", errors[0])
            self.assertIn("''", errors[0])

    def test_parse_xml_service_fields_error_name_too_long(self):
        long_boi = "this name is really long, why? Why would anyone prepare a medical " \
                   "service with a name so long, it doesn't make any sense..."
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>CODE</ServiceCode>
                    <ServiceName>{long_boi}</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>D</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>S</ServiceCategory>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertFalse(raw_services)
            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("name is invalid", errors[0])
            self.assertIn(long_boi, errors[0])

    def test_parse_xml_service_fields_error_unknown_type(self):
        unknown_type = "O"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>P_MAJ</ServiceCode>
                    <ServiceName>Service of type P in capital letter</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>d</ServiceLevel>
                    <ServicePrice>9.87</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>P_MIN</ServiceCode>
                    <ServiceName>Service of type P in small letter</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>H</ServiceLevel>
                    <ServicePrice>8.76</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>C_MAJ</ServiceCode>
                    <ServiceName>Service of type C in capital letter</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>h</ServiceLevel>
                    <ServicePrice>7.65</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>C_MIN</ServiceCode>
                    <ServiceName>Service of type C in small letter</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>S</ServiceLevel>
                    <ServicePrice>6.54</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>O_MIN</ServiceCode>
                    <ServiceName>Service of unknown type</ServiceName>
                    <ServiceType>{unknown_type}</ServiceType>
                    <ServiceLevel>s</ServiceLevel>
                    <ServicePrice>5.43</ServicePrice>
                    <ServiceCareType>B</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertTrue(raw_services)
            self.assertEqual(len(raw_services), 4)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("type is invalid", errors[0])
            self.assertIn(unknown_type, errors[0])

    def test_parse_xml_service_fields_error_unknown_care_type(self):
        unknown_care_type = "K"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>I_MAJ</ServiceCode>
                    <ServiceName>Service of care_type I in capital letter</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>V</ServiceLevel>
                    <ServicePrice>9.87</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>s</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>I_MIN</ServiceCode>
                    <ServiceName>Service of care_type I in small letter</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>v</ServiceLevel>
                    <ServicePrice>8.76</ServicePrice>
                    <ServiceCareType>i</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>D</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>O_MAJ</ServiceCode>
                    <ServiceName>Service of care_type O in capital letter</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>D</ServiceLevel>
                    <ServicePrice>7.65</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>d</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>O_MIN</ServiceCode>
                    <ServiceName>Service of care_type O in small letter</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>d</ServiceLevel>
                    <ServicePrice>6.54</ServicePrice>
                    <ServiceCareType>o</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>A</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>B_MAJ</ServiceCode>
                    <ServiceName>Service of care_type B in capital letter</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>H</ServiceLevel>
                    <ServicePrice>5.43</ServicePrice>
                    <ServiceCareType>B</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>a</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>B_MIN</ServiceCode>
                    <ServiceName>Service of care_type B in small letter</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>h</ServiceLevel>
                    <ServicePrice>4.32</ServicePrice>
                    <ServiceCareType>b</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>H</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
                <Service>
                    <ServiceCode>K_MAJ</ServiceCode>
                    <ServiceName>Service of unknown care_type</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>S</ServiceLevel>
                    <ServicePrice>3.21</ServicePrice>
                    <ServiceCareType>{unknown_care_type}</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory>h</ServiceCategory>
                    <ServiceFrequency>7</ServiceFrequency>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertTrue(raw_services)
            self.assertEqual(len(raw_services), 6)

            self.assertTrue(errors)
            self.assertEqual(len(errors), 1)
            self.assertIn("care type is invalid", errors[0])
            self.assertIn(unknown_care_type, errors[0])

    def test_parse_xml_service_fields_mixed_errors_and_success(self):
        code_ok_1 = "CODE_1"
        code_ok_2 = "CODE_2"
        code_ok_3 = "CODE_3"

        long_name = "this name is really long, why? Why would anyone prepare a medical service " \
                    "with a name so long, it doesn't make any sense..."
        unknown_care_type = "K"
        xml = f"""
            <Services>
                <Service>
                    <ServiceCode>CODE</ServiceCode>
                    <ServiceName>{long_name}</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>s</ServiceLevel>
                    <ServicePrice>1234</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_ok_1}</ServiceCode>
                    <ServiceName>Valid Service 1 - no error</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServiceLevel>V</ServiceLevel>
                    <ServicePrice>9.87</ServicePrice>
                    <ServiceCareType>I</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>K_MAJ</ServiceCode>
                    <ServiceName>Service of unknown care_type</ServiceName>
                    <ServiceType>c</ServiceType>
                    <ServiceLevel>v</ServiceLevel>
                    <ServicePrice>3.21</ServicePrice>
                    <ServiceCareType>{unknown_care_type}</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_ok_2}</ServiceCode>
                    <ServiceName>Valid Service 2 - no error</ServiceName>
                    <ServiceType>P</ServiceType>
                    <ServiceLevel>D</ServiceLevel>
                    <ServicePrice>555.55</ServicePrice>
                    <ServiceCareType>b</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceCode>{code_ok_3}</ServiceCode>
                    <ServiceName>Valid Service 3 - no error</ServiceName>
                    <ServiceType>p</ServiceType>
                    <ServiceLevel>d</ServiceLevel>
                    <ServicePrice>489.54</ServicePrice>
                    <ServiceCareType>O</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
                <Service>
                    <ServiceName>Service without code</ServiceName>
                    <ServiceType>C</ServiceType>
                    <ServicePrice>3700.88</ServicePrice>
                    <ServiceLevel>H</ServiceLevel>
                    <ServiceCareType>o</ServiceCareType>
                    <ServiceMaleCategory>1</ServiceMaleCategory>
                    <ServiceFemaleCategory>1</ServiceFemaleCategory>
                    <ServiceAdultCategory>1</ServiceAdultCategory>
                    <ServiceMinorCategory>1</ServiceMinorCategory>
                    <ServiceCategory/>
                    <ServiceFrequency/>
                </Service>
            </Services>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)

            raw_services, errors = parse_xml_services(et)

            self.assertTrue(raw_services)
            self.assertEqual(len(raw_services), 3)
            self.assertIn(code_ok_1, raw_services[0]["code"])
            self.assertIn(code_ok_2, raw_services[1]["code"])
            self.assertIn(code_ok_3, raw_services[2]["code"])

            self.assertTrue(errors)
            self.assertEqual(len(errors), 3)
            self.assertIn("name is invalid", errors[0])
            self.assertIn(long_name, errors[0])
            self.assertIn("care type is invalid", errors[1])
            self.assertIn(unknown_care_type, errors[1])
            self.assertIn("Service is missing one of", errors[2])


class UploadServicesParseOptionalFieldsTestCase(TestCase):

    def test_parse_optional_service_fields_all_empty(self):
        xml = f"""
            <Service>
                <ServiceCategory/>
                <ServiceFrequency/>
            </Service>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_service_fields(root, "EMPTY")

            self.assertFalse(optional_fields)
            self.assertFalse(error)

    def test_parse_optional_service_fields_all_filled(self):
        xml = f"""
            <Service>
                <ServiceCategory>O</ServiceCategory>
                <ServiceFrequency>5</ServiceFrequency>
            </Service>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_service_fields(root, "FILLED")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 2)
            self.assertIn("frequency", optional_fields)
            self.assertIn("category", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_service_fields_filled_and_empty_frq(self):
        xml = f"""
            <Service>
                <ServiceCategory/>
                <ServiceFrequency>78</ServiceFrequency>
            </Service>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_service_fields(root, "1/2")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 1)
            self.assertIn("frequency", optional_fields)
            self.assertNotIn("category", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_service_fields_filled_and_empty_cat(self):
        xml = f"""
            <Service>
                <ServiceCategory>o</ServiceCategory>
                <ServiceFrequency/>
            </Service>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            optional_fields, error = parse_optional_service_fields(root, "1/3")

            self.assertTrue(optional_fields)
            self.assertEqual(len(optional_fields), 1)
            self.assertIn("category", optional_fields)
            self.assertNotIn("frequency", optional_fields)
            self.assertFalse(error)

    def test_parse_optional_service_fields_error_frequency(self):
        xml = f"""
            <Service>
                <ServiceCategory>V</ServiceCategory>
                <ServiceFrequency>5.55</ServiceFrequency>
            </Service>
        """.encode("utf8")
        with TemporaryFile() as tf:
            tf.write(xml)
            tf.seek(0)
            et = ElementTree.parse(tf)
            root = et.getroot()

            code = "ERROR_FRQ"
            optional_fields, error = parse_optional_service_fields(root, code)

            self.assertNotIn("frequency", optional_fields)
            self.assertTrue(error)
            self.assertIn("frequency is invalid", error)
            self.assertIn(code, error)


class UploadServicesTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username="testServicesAdmin")

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT
        existing_code = "A1"
        raw_services = [
            {
                "code": "CODE_1",
                "name": "Valid Service 1 - no error",
            },
            {
                "code": "CODE_2",
                "name": "Valid Service 2 - no error",
            },
            {
                "code": existing_code,
                "name": "Invalid Service - code already exists",
            }
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors

        expected_errors = [
            f"Service '{existing_code}' already exists"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=2,
            updated=0,
            deleted=0,
        )
        result = upload_services(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_update_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_UPDATE
        existing_code_1 = "A1"
        existing_code_2 = "I117"
        error_code = "ERROR"
        raw_services = [
            {
                "code": existing_code_1,
                "name": "Valid Service 1 - no error",
            },
            {
                "code": error_code,
                "name": "Invalid Service - code doesn't exists",
            },
            {
                "code": existing_code_2,
                "name": "Valid Service 2 - no error",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors

        expected_errors = [
            f"Service '{error_code}' does not exist"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=0,
            updated=2,
            deleted=0,
        )
        result = upload_services(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert_update_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT_UPDATE
        raw_services = [
            {
                "code": "CODE_1",
                "name": "New Service 1",
            },
            {
                "code": "CODE_2",
                "name": "New Service 2",
            },
            {
                "code": "G1",
                "name": "Existing Service 1",
            },
            {
                "code": "CODE_3",
                "name": "New Service 3",
            },
            {
                "code": "G34",
                "name": "Existing Service 2",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors

        expected = UploadResult(
            errors=errors,
            sent=5,
            created=3,
            updated=2,
            deleted=0,
        )
        result = upload_services(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert_update_delete_dry_run(self, mock_parsing):
        dry_run = True
        strategy = STRATEGY_INSERT_UPDATE_DELETE
        raw_services = [
            {
                "code": "CODE_1",
                "name": "New Service 1",
            },
            {
                "code": "CODE_2",
                "name": "New Service 2",
            },
            {
                "code": "A1",
                "name": "Existing Service 1",
            },
            {
                "code": "CODE_3",
                "name": "New Service 3",
            },
            {
                "code": "G60",
                "name": "Existing Service 2",
            },
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors

        before_total_services = Service.objects.filter(*filter_validity()).count()
        expected_deleted = before_total_services - 2

        expected = UploadResult(
            errors=errors,
            sent=5,
            created=3,
            updated=2,
            deleted=expected_deleted,
        )
        result = upload_services(self.admin_user, "xml", strategy, dry_run)

        self.assertEqual(expected, result)

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert(self, mock_parsing):
        # setup - preparing data that will be inserted
        dry_run = False
        strategy = STRATEGY_INSERT
        existing_code = "A1"
        new_code_1 = "CODE_1"
        new_code_2 = "CODE_2"
        raw_services = [
            {
                "code": new_code_1,
                "name": "Valid Service 1 - no error",
                "type": "P",
                "level": "S",
                "price": 489.54,
                "care_type": "O",
                "patient_category": 15,
                "category": "H",
                "frequency": 5,
            },
            {
                "code": new_code_2,
                "name": "Valid Service 2 - no error",
                "type": "C",
                "level": "s",
                "price": 499.54,
                "care_type": "O",
                "patient_category": 5,
                "category": "c",
                "frequency": 57,
            },
            {
                "code": existing_code,
                "name": "Invalid Service - code already exists",
                "type": "p",
                "level": "V",
                "price": 599.54,
                "care_type": "B",
                "patient_category": 7,
                "category": "O",
                "frequency": 7,
            }
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors

        expected_errors = [
            f"Service '{existing_code}' already exists"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=2,
            updated=0,
            deleted=0,
        )
        total_services_before = Service.objects.filter(*filter_validity()).count()

        # Inserting
        result = upload_services(self.admin_user, "xml", strategy, dry_run)
        total_services_after = Service.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_services_before + 2, total_services_after)

        # Making sure the new services don't stay in the DB
        new_service1 = Service.objects.get(code=new_code_1)
        new_service1.delete()
        new_service2 = Service.objects.get(code=new_code_2)
        new_service2.delete()

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_update(self, mock_parsing):
        # setup - creating services that will be updated
        new_code_1 = "CODE_1"
        old_name_1 = "new Service old name 0001"
        new_service_1_props = {
            "code": new_code_1,
            "name": old_name_1,
        }
        create_test_service(category="D", custom_props=new_service_1_props)

        new_code_2 = "CODE_2"
        old_name_2 = "new Service old name 0002"
        new_service_2_props = {
            "code": new_code_2,
            "name": old_name_2,
        }
        create_test_service(category="D", custom_props=new_service_2_props)

        # setup - preparing values used for the update
        dry_run = False
        strategy = STRATEGY_UPDATE
        new_name_1 = "new Service new name 0001"
        new_name_2 = "new Service new name 0002"
        non_existing_code = "ERROR"
        raw_services = [
            {
                "code": new_code_1,
                "name": new_name_1,
                "type": "c",
                "level": "v",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "category": "D",
            },
            {
                "code": new_code_2,
                "name": new_name_2,
                "type": "P",
                "level": "D",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
                "frequency": 1,
            },
            {
                "code": non_existing_code,
                "name": "error - this can't be updated",
                "type": "p",
                "level": "d",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
            }
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors
        expected_errors = [
            f"Service '{non_existing_code}' does not exist"
        ]
        expected = UploadResult(
            errors=expected_errors,
            sent=3,
            created=0,
            updated=2,
            deleted=0,
        )
        total_services_before = Service.objects.filter(*filter_validity()).count()

        # update
        result = upload_services(self.admin_user, "xml", strategy, dry_run)
        total_services_after = Service.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_services_before, total_services_after)

        # Making sure the names have been updated + deleting the new services to make sure they don't stay in the DB
        db_new_service_1 = Service.objects.get(code=new_code_1, validity_to=None)
        self.assertEqual(db_new_service_1.name, new_name_1)
        db_new_service_1.delete()

        db_new_service_2 = Service.objects.get(code=new_code_2, validity_to=None)
        self.assertEqual(db_new_service_2.name, new_name_2)
        db_new_service_2.delete()

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert_update(self, mock_parsing):
        # setup - creating services that will be updated
        update_code_1 = "U_1"
        old_name_1 = "update Service old name 0001"
        update_service_1_props = {
            "code": update_code_1,
            "name": old_name_1,
        }
        create_test_service(category="D", custom_props=update_service_1_props)

        update_code_2 = "U_2"
        old_name_2 = "update Service old name 0002"
        update_service_2_props = {
            "code": update_code_2,
            "name": old_name_2,
        }
        create_test_service(category="D", custom_props=update_service_2_props)

        # setup - preparing values used for the update
        dry_run = False
        strategy = STRATEGY_INSERT_UPDATE
        new_name_1 = "update Service new name 0001"
        new_name_2 = "update Service new name 0002"
        insert_code_1 = "I_1"
        insert_code_2 = "I_2"
        raw_services = [
            {
                "code": insert_code_1,
                "name": "insert Service 1",
                "type": "C",
                "level": "H",
                "price": 65.54,
                "care_type": "b",
                "patient_category": 15,
                "category": "d",
            },
            {
                "code": update_code_1,
                "name": new_name_1,
                "type": "c",
                "level": "h",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "category": "D",
                "frequency": 6,
            },
            {
                "code": update_code_2,
                "name": new_name_2,
                "type": "P",
                "level": "S",
                "price": 66.54,
                "care_type": "I",
                "patient_category": 15,
            },
            {
                "code": insert_code_2,
                "name": "insert Service 2",
                "type": "p",
                "level": "s",
                "price": 60.54,
                "care_type": "b",
                "patient_category": 15,
            },
        ]
        errors = []
        mock_parsing.return_value = raw_services, errors
        expected = UploadResult(
            errors=errors,
            sent=4,
            created=2,
            updated=2,
            deleted=0,
        )
        total_services_before = Service.objects.filter(*filter_validity()).count()

        # insert-update
        result = upload_services(self.admin_user, "xml", strategy, dry_run)
        total_services_after = Service.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        self.assertEqual(total_services_before + 2, total_services_after)

        # Making sure the names have been updated + deleting the update services to make sure they don't stay in the DB
        db_update_service_1 = Service.objects.get(code=update_code_1, validity_to=None)
        self.assertEqual(db_update_service_1.name, new_name_1)
        db_update_service_1.delete()

        db_update_service_2 = Service.objects.get(code=update_code_2, validity_to=None)
        self.assertEqual(db_update_service_2.name, new_name_2)
        db_update_service_2.delete()

        # Also deleting the insert services to make sure they don't stay in the DB
        db_insert_service_1 = Service.objects.get(code=insert_code_1, validity_to=None)
        db_insert_service_1.delete()
        db_insert_service_2 = Service.objects.get(code=insert_code_2, validity_to=None)
        db_insert_service_2.delete()

    @patch("tools.services.parse_xml_services")
    def test_upload_services_multiple_insert_update_delete(self, mock_parsing):
        # setup - fetching initial DB services in order not to delete them
        services = Service.objects.filter(*filter_validity()).all()
        services_to_not_delete = []
        for service in services:
            service_dict = service.__dict__
            service_dict.pop("id")
            service_dict.pop("uuid")
            service_dict.pop("_state")
            service_dict.pop("validity_from")
            service_dict.pop("validity_to")
            service_dict.pop("legacy_id")
            services_to_not_delete.append(service_dict)

        # setup - creating services that will be updated & deleted
        update_code = "U_1"
        old_name = "update Service old name 0001"
        update_service_props = {
            "code": update_code,
            "name": old_name,
        }
        create_test_service(category="D", custom_props=update_service_props)

        delete_code = "D_1"
        delete_service_props = {
            "code": delete_code,
            "name": "Service do be deleted",
        }
        create_test_service(category="D", custom_props=delete_service_props)

        # setup - preparing values used for the tests
        dry_run = False
        strategy = STRATEGY_INSERT_UPDATE_DELETE
        update_new_name = "update Service new name 0001"
        insert_code_1 = "I_1"
        insert_code_2 = "I_2"
        new_services = [
            {
                "code": insert_code_1,
                "name": "insert Service 1",
                "type": "p",
                "level": "V",
                "price": 65.54,
                "care_type": "b",
                "patient_category": 15,
                "category": "s",
            },
            {
                "code": update_code,
                "name": update_new_name,
                "type": "c",
                "level": "v",
                "price": 66.54,
                "care_type": "i",
                "patient_category": 15,
                "category": "S",
                "frequency": 6,
            },
            {
                "code": insert_code_2,
                "name": "insert Service 2",
                "type": "c",
                "level": "D",
                "price": 60.54,
                "care_type": "b",
                "patient_category": 15,
            },
        ]
        raw_services = new_services + services_to_not_delete
        errors = []
        mock_parsing.return_value = raw_services, errors

        total_updated = len(services_to_not_delete) + 1
        total_sent = total_updated + 2
        expected = UploadResult(
            errors=errors,
            sent=total_sent,
            created=2,
            updated=total_updated,
            deleted=1,
        )
        total_services_before = Service.objects.filter(*filter_validity()).count()

        # insert update delete
        result = upload_services(self.admin_user, "xml", strategy, dry_run)
        total_services_after = Service.objects.filter(*filter_validity()).count()

        self.assertEqual(expected, result)
        # 2 inserts and 1 deletion
        self.assertEqual(total_services_before + 1, total_services_after)

        with self.assertRaises(ObjectDoesNotExist):
            Service.objects.get(code=delete_code, validity_to=None)

        # Making sure the name has been updated + deleting the services to make sure they don't stay in the DB
        db_update_service = Service.objects.get(code=update_code, validity_to=None)
        self.assertEqual(db_update_service.name, update_new_name)
        db_update_service.delete()

        db_insert_service_1 = Service.objects.get(code=insert_code_1, validity_to=None)
        db_insert_service_1.delete()
        db_insert_service_2 = Service.objects.get(code=insert_code_2, validity_to=None)
        db_insert_service_2.delete()



# Generated by Django 3.0.14 on 2021-12-08 11:17
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("tools", "0001_initial"),
    ]

    operations = [
        migrations.RunSQL(
            "ALTER TABLE tblExtracts ADD ExtractFile nvarchar(255) NULL"
            if settings.MSSQL else
            'ALTER TABLE "tblExtracts" ADD "ExtractFile" varchar(255)',
            reverse_sql='ALTER TABLE "tblExtracts" DROP "ExtractFile"',
        ),
    ]
